#include <AccelStepper.h>
#include <MultiStepper.h>
#include <Servo.h>
#include <Wire.h>
#include "mpu9250.h"
#include "ewma.h"


/*
  Using a Pololu A4988 or StepStick Stepper Motor Driver with FabScan-Shield.
*/

            //driver: 1 | 2 | 3  | 4
#define EN_PIN_3   11 //2 | 5 | 11 | A0
#define EN_PIN_2    5
#define EN_PIN_1    2
#define STEP_PIN    3 //3 | 6 | 12 | A1
#define DIR_PIN     4 //4 | 7 | 13 | A2
#define MS_PIN     A5 //HIGH for 16 microstepping, LOW for no microstepping
#define IRPROBE_PIN 9
#define SERVO_PIN   8
#define HEAD_LIMIT_PIN A0
#define DRAWER_PULLUP_PIN A1
#define DRAWER_LIMIT_PIN A2

float ewma(int input);
int irProbe();
void serialParseCmd(char* buf);

AccelStepper stepper1(1, 3, 4);
AccelStepper stepper2(1, 6, 7);
AccelStepper stepper3(1, 12, 13);
AccelStepper* stepper[3];
Servo servo;

CMPU9250 imu;
CEWMA vectorFilt, gyroFilt;

int g_headMaxSpeed = 2800;
int g_headMaxAccel = 4000;
int g_headSurfaceHeight = 0;

int g_signalUnload = 0;
int g_signalPutDown = 0;
int g_signalPickup = 0;
int g_signalLoad = 0;

int g_waitForSignal = 0;

float g_vector, g_gyro;

void setup()
{
  Serial.begin(9600);

  //Steppers enable low active
  pinMode(EN_PIN_1, OUTPUT);
  digitalWrite(EN_PIN_1, LOW); 
  pinMode(EN_PIN_2, OUTPUT);
  digitalWrite(EN_PIN_2, LOW); 
  pinMode(EN_PIN_3, OUTPUT);
  digitalWrite(EN_PIN_3, LOW); 
  
  //Set the direction for the irprobe input pin
  pinMode(IRPROBE_PIN, INPUT);
  
  stepper1.setMaxSpeed(3000.0);  
  stepper1.setAcceleration(1000.0);  
  stepper2.setMaxSpeed(40.0);  
  stepper2.setAcceleration(8.0);
  stepper3.setMaxSpeed(g_headMaxSpeed);  
  stepper3.setAcceleration(g_headMaxAccel);   

  stepper[0] = &stepper1;
  stepper[1] = &stepper2;
  stepper[2] = &stepper3;

  //Servo on pin 8
  servo.attach(SERVO_PIN);
  servo.write(60);

  //Drawer limit switch on pins A1 A2
  pinMode(DRAWER_PULLUP_PIN, OUTPUT);
  digitalWrite(DRAWER_PULLUP_PIN, LOW); //One side of switch is high
  pinMode(DRAWER_LIMIT_PIN, INPUT_PULLUP);

  stepper[1]->setCurrentPosition(0);

  imu.init();
  imu.calibrateDownVector();
  vectorFilt.init(0.05);
  gyroFilt.init(0.01);
}

void readImu()
{
  float vector, gyro;
  imu.read(&vector, &gyro);
  vector = abs(vector);
  gyro = abs(gyro);
  vectorFilt.run(&vector, &g_vector);
  gyroFilt.run(&gyro, &g_gyro);
}

bool checkImuValues()
{
  bool rv = true;

  readImu();

  if(true) {
    Serial.print("IMU: ");
    Serial.print(g_vector);
    Serial.print(" | ");
    Serial.println(g_gyro);
  }

  //if(abs(g_vector) > 410) rv = false;
  if(abs(g_gyro)   > 5)   rv = false;

  return rv;
}

void waitForImu()
{
  vectorFilt.set(999);
  gyroFilt.set(9);

  while(checkImuValues() == false) {};
}

void waitForStepper()
{
  bool wait;
  do {
    wait = false;
    for(int i=0; i<3; i++) {
      if(stepper[i]->run()) {
        wait = true;
      }
    }
        
  } while(wait);
}

void servoEject()
{
  servo.write(150);
  delay(800);
  servo.write(60);
}

bool headLimitSwitch() 
{
  int val = analogRead(HEAD_LIMIT_PIN);
  float filteredVal = ewma(val);
  //Serial.println(filteredVal);
  if(filteredVal > 1.8) {
    return true;
  } else {
    return false;
  }
}

float ewma(int input)
{
  static int oldInput = 0;
  static float q = 0.0f;
  static float alpha = 0.02f;

  float deriv=0.0f;
  if(oldInput != 0) { 
    deriv = input - oldInput;
  }
  oldInput = input;

  q=deriv*alpha + q*(1-alpha);

  return q;
}

int headHoming()
{
  int rv = 0;
  stepper[2]->setMaxSpeed(100);
  stepper[2]->move(170);
  while(stepper[2]->run()) {};
  while(headLimitSwitch() == false) {
    stepper[2]->move(-1);
    while(stepper[2]->run()) {};
  }
  
  stepper[2]->setCurrentPosition(0);
  stepper[2]->moveTo(10);
  while(stepper[2]->run()) {};
  stepper[2]->setMaxSpeed(g_headMaxSpeed);
  
  return rv;
}

void headUp()
{
  stepper[2]->moveTo(0);
}

void headDroppingPosition()
{
  stepper[2]->moveTo(200);
  while(stepper[2]->run()) {};
}

int headFindSurface()
{
  AccelStepper* s = stepper[2];

  while(irProbe() == 0) {
    s->move(10);
    if(s->currentPosition() > 4000) {
      return -1;
    }
    s->run();
  }
  s->moveTo(s->currentPosition());
  while(s->run()) {};

  waitForImu();

  s->setMaxSpeed(100);
  s->move(-200);
  while(s->run()) {};

  while(irProbe() == 0) {
    s->move(1);
    if(s->currentPosition() > 4000) {
      return -2;
    }
    s->run();
    delay(5);
  }
  s->setMaxSpeed(g_headMaxSpeed);

  return s->currentPosition();
}

int irProbe()
{
  int rv = digitalRead(IRPROBE_PIN);
  static int oldIrProbe = 0;
  
  if(oldIrProbe != rv) {
    oldIrProbe = rv;
  }
  return rv;
}

int g_drawerIsHomedYet = 0;

int drawerLimitSwitch()
{
  static int lastPinState = 1;
  static int buttonState = 1;
  static unsigned long lastDebounceTime = 0;
  int pinState = digitalRead(DRAWER_LIMIT_PIN) ^ 1;
  const int debounceDelay = 10;

  //If pinState changed
  if(pinState != lastPinState) {
    //...save time in lastDebounceTime
    lastDebounceTime = millis();
  }

  //if enough time has passed
  if( (millis() - lastDebounceTime) > debounceDelay) {
    if(pinState != buttonState) {
      buttonState = pinState;
    }
  }

  lastPinState = pinState;

  return buttonState;
}

void drawerHoming() 
{
  AccelStepper* s = stepper[0];
  int speed = s->maxSpeed();

  if(s->currentPosition() < 1700) {
    s->move(50);
    while(s->run()) {};
  }
  for(int i=0; i<3; i++) {
    drawerLimitSwitch();
    delay(8);
  }
  while(!drawerLimitSwitch()) {
    s->move(-20);
    while(s->run()) {};
  }
  s->setMaxSpeed(100);
  s->move(200);
  while(s->run()) {};
  for(int i=0; i<3; i++) {
    drawerLimitSwitch();
    delay(8);
  }
  while(!drawerLimitSwitch()) {
    s->move(-1);
    while(s->run()) {};
  }

  s->setMaxSpeed(speed);
  s->setCurrentPosition(0);
  g_drawerIsHomedYet = 1;
}

void drawerIn()
{
  if(!g_drawerIsHomedYet) {
    drawerHoming();
  }
  AccelStepper* s = stepper[0];
  s->moveTo(400);
}

void drawerOut()
{
  if(!g_drawerIsHomedYet) {
    drawerHoming();
  }
  AccelStepper* s = stepper[0];
  s->moveTo(1710);
}

void serialIo(void)
{
  const int bufSize=8;
  static char buf[bufSize+1];
  static int bufPos = 0;
  while(Serial.available()) {
    unsigned char c = Serial.read();

    if(c == 0x7F) { //Backspace
      bufPos--;
      if(bufPos < 0) bufPos=0;
    } else {
      buf[bufPos] = c;
      bufPos++;
      if(bufPos > bufSize-1) bufPos = bufSize-1;
    }

    //Clear line
    Serial.write('\r');
    for(int i=0; i<bufSize; i++) {
      Serial.write(' ');
    }
    Serial.write('\r');
    //Send buf
    buf[bufPos] = '\0';
    Serial.print(buf);

    if(c == '\r') {
      Serial.print("\r\n");
      serialParseCmd(buf);
      bufPos = 0;
    }
  }

  if(g_waitForSignal == 1 &&
     g_signalUnload == 0 &&
     g_signalLoad == 0 &&
     g_signalPickup == 0 &&
     g_signalPutDown == 0) {
    g_waitForSignal = 0;
    Serial.println("done\r\n");
  }
}

void serialParseCmd(char* buf) 
{
  int cmd = atoi(buf+0);
  int param = atoi(buf+2);

  if(cmd > 0 && cmd < 4) {
    stepper[cmd - 1]->moveTo(param);
  }
  
  if(cmd == 4) {
    servoEject();
  }
  if(cmd == 5) {
    headHoming();
  }

  if(cmd == 6) {
    drawerOut();
    Serial.println("done\r\n");
  }

  if(cmd == 7) {
    drawerIn();
    Serial.println("done\r\n");
  }

  if(cmd == 8) {
    g_headSurfaceHeight = headFindSurface();
    Serial.print("Surface height is:");
    Serial.println(g_headSurfaceHeight);
  }

  if(cmd == 10) {
    g_signalPickup = 1;
    g_waitForSignal = 1;
  }

  if(cmd == 11) {
    g_signalLoad = 1;
    g_waitForSignal = 1;
  }

  if(cmd == 12) {
    g_signalUnload = 1;
    g_waitForSignal = 1;
  }

  if(cmd == 13) {
    g_signalPutDown = 1;
    g_waitForSignal = 1;
  }
}

void offsetIn()
{
  AccelStepper* s = stepper[1];
  s->moveTo(40);
}

void offsetOut()
{
  AccelStepper* s = stepper[1];
  s->moveTo(0);
}

typedef enum 
{
  EST_start=0,
  EST_homeHead,
  EST_findSurfaceHeight,
  EST_checkIfFinished,
  EST_pickUp,
  EST_lift,
  EST_check,
  EST_success
} EN_StatePickUp_t;

EN_StatePickUp_t g_statePickUp;
int g_pickUpCounter = 0;
int g_pickUpFailCounter = 0;
int g_pickUpCounterSinceHomeing = 0;
const int g_surfaceToPickupDistance = 100;
const int g_finishedSurfaceHeight = 2144+g_surfaceToPickupDistance-10;
int g_checkSurfaceBeforeFinish = 0;

void stateMachinePickUp()
{
  int currentSurfaceHeight = g_headSurfaceHeight + 
                             g_surfaceToPickupDistance + 
                             18*g_pickUpCounterSinceHomeing;
  switch(g_statePickUp) {
    case EST_start:
      if(g_signalPickup > 0) {
        drawerIn();
        g_statePickUp = EST_homeHead;
      }
      break;

    case EST_homeHead:
      headHoming();
      waitForImu();
      g_statePickUp = EST_findSurfaceHeight;
      break;

    case EST_findSurfaceHeight:
      if(g_headSurfaceHeight == 0 || 
         g_pickUpFailCounter > 2 || 
         g_checkSurfaceBeforeFinish != 0) {

        g_headSurfaceHeight = headFindSurface();
        g_pickUpCounterSinceHomeing = 0;
        stepper[2]->moveTo(0);
        waitForStepper();
      }
      g_statePickUp = EST_checkIfFinished;

      break;

    case EST_checkIfFinished:
      if(currentSurfaceHeight >= g_finishedSurfaceHeight) {
        //First try...
        if(g_checkSurfaceBeforeFinish == 0) {
          g_checkSurfaceBeforeFinish = 1;
          g_statePickUp = EST_homeHead;
        } else {
          g_signalPickup = -1;
          Serial.println("finished");
          g_statePickUp = EST_start;
        }
      } else {
        g_checkSurfaceBeforeFinish = 0;
        g_statePickUp = EST_pickUp;
      }
      break;

    case EST_pickUp:
      waitForImu();
      stepper[2]->moveTo(currentSurfaceHeight);
      g_statePickUp = EST_lift;
      break;

    case EST_lift:
      delay(1000);
      stepper[2]->moveTo(0);
      g_statePickUp = EST_check;
      break;

    case EST_check:
      if(irProbe() == 1) {
        g_pickUpCounter++;
        g_pickUpCounterSinceHomeing++;
        g_statePickUp = EST_success;
        delay(2000);
      } else {
        waitForImu();
        if(g_pickUpFailCounter++ > 2) {
          g_statePickUp = EST_homeHead;
        } else {
          g_statePickUp = EST_pickUp;
        }
      }
      break;

    case EST_success:
      g_pickUpFailCounter = 0;
      g_signalPickup = 0;
      g_statePickUp = EST_start;
      break;
  }
}

typedef enum 
{
  ESU_start,
  ESU_headHoming,
  ESU_pickup,
  ESU_lift,
  ESU_check,
  ESU_success
} EN_stateUnload_t;

EN_stateUnload_t g_stateUnload;


void stateMachineUnload()
{
  switch(g_stateUnload) {
    case ESU_start:
      if(g_signalUnload > 0) {
        g_stateUnload = ESU_headHoming;
      }
      break;

    case ESU_headHoming:
      offsetIn();
      waitForStepper();
      headHoming();
      g_stateUnload = ESU_pickup;
      break;

    case ESU_pickup:
      waitForImu();
      stepper[2]->moveTo(460);
      waitForStepper();
      delay(500);
      g_stateUnload = ESU_lift;
      break;

    case ESU_lift:
      stepper[2]->moveTo(360);
      offsetOut();
      waitForStepper();
      stepper[2]->moveTo(0);
      waitForStepper();
      g_stateUnload = ESU_check;
      break;

    case ESU_check:
      waitForImu();
      if(irProbe() == 1) {
        g_stateUnload = ESU_success;
      } else {
        g_stateUnload = ESU_headHoming;
      }
      break;

    case ESU_success:
      g_signalUnload = 0;
      g_stateUnload = ESU_start;
      break;
  }
}

typedef enum {
  EPD_start,
  EPD_moveDrawer,
  EPD_moveHeadDown,
  EPD_eject,
  EPD_moveHeadUp,
  EPD_success
} EN_statePutDown_t;

EN_statePutDown_t g_statePutDown;
int g_putDownSurfaceLevel = 1500;
int g_putDownCounter = 0;
int g_putDownCounterSinceHoming = 0;

void stateMachinePutDown()
{
  switch(g_statePutDown) {
    case EPD_start:
      if(g_signalPutDown > 0) {
        g_statePutDown = EPD_moveDrawer;
      }
      break;

    case EPD_moveDrawer:
      drawerOut();
      g_statePutDown = EPD_moveHeadDown;
      break;

    case EPD_moveHeadDown:
      stepper[2]->moveTo(g_putDownSurfaceLevel-30*g_putDownCounterSinceHoming);
      g_statePutDown = EPD_eject;
      break;

    case EPD_eject:
      servoEject();
      g_statePutDown = EPD_moveHeadUp;
      break;

    case EPD_moveHeadUp:
      stepper[2]->moveTo(0);
      g_statePutDown = EPD_success;
      break;

    case EPD_success:
      g_putDownCounter++;
      g_putDownCounterSinceHoming++;
      g_signalPutDown = 0;
      g_statePutDown = EPD_start;
      break;
  }
}

typedef enum {
  ESL_start,
  ESL_homeHead,
  ESL_prepare,
  ESL_drop,
  ESL_success
} EN_stateLoad_t;

EN_stateLoad_t g_stateLoad;

void stateMachineLoad()
{
  switch(g_stateLoad) {
    case ESL_start:
      if(g_signalLoad > 0) {
        g_stateLoad = ESL_homeHead;
      }
      break;

    case ESL_homeHead:
      headHoming();
      g_stateLoad = ESL_prepare;
      break;

    case ESL_prepare:
      offsetIn();
      waitForStepper();
      waitForImu();
      stepper[2]->moveTo(200);
      g_stateLoad = ESL_drop;
      break;

    case ESL_drop:
      servoEject();
      offsetOut();
      stepper[2]->moveTo(0);
      g_stateLoad = ESL_success;
      break;

    case ESL_success:
      g_signalLoad = 0;
      g_stateLoad = ESL_start;
      break;
  }
}

void loop()
{
  serialIo();    
  waitForStepper();
  stateMachinePickUp();
  stateMachineUnload();
  stateMachinePutDown();
  stateMachineLoad();
}

